{****************************************************}
{  Simple reservoir model exported functions.        }
{  Copyright (c) 2007 by                             }
{  National Technical University of Athens           }
{  Authors, Stefanos Kozanis, Andreas Efstratiadis,  }
{  Giannis Liagos.                                   }
{****************************************************}

unit expfuncs;

interface

uses SysUtils, Classes, IniFiles, Contnrs, Dates, ts, reservoir;

function Initialize(filepath : ShortString): Boolean ; cdecl ;

function PerformTimeStep() : Boolean ; cdecl ;

function Finish() : Boolean ; cdecl ;

function Dispose() : Boolean ; cdecl ;

function GetStorage(var Strg : Double) : Boolean ; cdecl ;

function GetModelID(var id : ShortString ) : Boolean ; cdecl ;

function GetModelDescription(var desc : ShortString) : Boolean ; cdecl ;

function GetTimeStepLength(var tstep : Double) : Boolean ; cdecl ;

function GetNumberOfTimeSteps(var tsteps : Integer) : Boolean ; cdecl ;

function GetInputTime(var InptTime : Double) : Boolean ; cdecl ;

function GetCurrentTime(var CurrTime : Double) : Boolean ; cdecl ;

function GetSimulationStartDate(var SimStDate : ShortString) : Boolean ; cdecl ;

function AddInFlow(infl : Double) : Boolean ; cdecl ;

function GetFlow(var flow : Double) : Boolean ; cdecl ;

procedure GetMessage(var n : Integer ; var MessageFromCore : ShortString) cdecl ;

function GetNumberOfMessages() : Integer ; cdecl ;

function RunSimulation: PChar; pascal;

function RunSimulationArgs(filepath : ShortString): PChar; pascal;

var
  RunInteractive: Boolean;

implementation

uses tsprocess, Math;

{This is the core engine for the simple reservoir model.
 If we want to use the engine as a standalone application we must also define a
 RunSimilation() function that runs a whole simulation of the model. This function
 is called by our application.
 The functions below (except RunSimulation) are necessary in order to implement the OpenMI standard Interface.
 cdecl is used because we want to access the functions with C# from the .NET enviroment}

{-------------}

{ Global variables... }
var
  Storage, PrevStorage,    {The storage of the reservoir}
  Inflow, Outflow: Real ; {Inflow: Input / WaterDemand : Output}
  Timestep: TTimeStep;
  TimeSteps : Integer ;
  Counter : Integer ;    //It counts the performed TimeSteps
{ For handling errors}
  Errors : TStringList ;
  NumberOfErrorMessages : Integer ;
{Data types}
  AReservoir: TReservoir;
  ATargetSet: TTargetSet;
  Stage, InitialStage, Spill: Real;
  ADateTime, StartDateTime: TDateTime;
  Rainfall, Evaporation, MiscInflowVolume,
  MiscOutflowVolume, Runoff, ResultingOutflow,
  SpillTimeseries, StageTimeseries, StorageTimeseries: TTimeseries;
  ResultingOutflowFile, SpillTimeseriesFile, StageTimeseriesFile,
  StorageTimeseriesFile: string;
  DataDirectory: string;
  OutputDirectory: string;
  ADateTimeList: TDateTimeList;
  RunoffInMM: Boolean;

{Some local functions...}

procedure FreeTimeseries;
begin
  Rainfall.Free;
  Evaporation.Free;
  MiscInflowVolume.Free;
  MiscOutflowVolume.Free;
  Runoff.Free;
  ResultingOutflow.Free;
  SpillTimeseries.Free;
  StageTimeseries.Free;
  StorageTimeseries.Free;
end;

procedure NilTimeseries;
begin
  Rainfall := nil;
  Evaporation := nil;
  MiscInflowVolume := nil;
  MiscOutflowVolume := nil;
  Runoff := nil;
  ResultingOutflow := nil;
  SpillTimeseries := nil;
  StageTimeseries := nil;
  StorageTimeseries := nil;
end;

{This function initializes the engine
 The files we use are specific}

function Initialize(filepath : ShortString): Boolean ; cdecl ;
var
  err : ShortString ;
  INI: TMemIniFile;
  ATimeseriesList: TObjectList;
  SavedDecimalSeparator: Char;

  procedure CreateTimeseries;
  begin
    Rainfall := TTimeseries.Create;
    Evaporation := TTimeseries.Create;
    MiscInflowVolume := TTimeseries.Create;
    MiscOutflowVolume := TTimeseries.Create;
    Runoff := TTimeseries.Create;
    ResultingOutflow := TTimeseries.Create;
    SpillTimeseries := TTimeseries.Create;
    StageTimeseries := TTimeseries.Create;
    StorageTimeseries := TTimeseries.Create;

    Rainfall.TimeStep := Timestep;
    Evaporation.TimeStep := Timestep;
    MiscInflowVolume.TimeStep := Timestep;
    MiscOutflowVolume.TimeStep := Timestep;
    Runoff.TimeStep := Timestep;
    ResultingOutflow.TimeStep := Timestep;
    SpillTimeseries.TimeStep := Timestep;
    StageTimeseries.TimeStep := Timestep;
    StorageTimeseries.TimeStep := Timestep;
  end;

  procedure LoadTimeseries;
    procedure LoadSingleTS(ATimeseries: TTimeseries; s: string);
    begin
      if s<>'' then ATimeseries.LoadFromFile(DataDirectory+s);
    end;
  begin
    LoadSingleTS(Rainfall, INI.ReadString('InputFiles', 'RainfallTimeseries', ''));
    LoadSingleTS(Evaporation, INI.ReadString('InputFiles', 'EvaporationTimeseries', ''));
    LoadSingleTS(MiscInflowVolume, INI.ReadString('InputFiles', 'InflowTimeseries', ''));
    LoadSingleTS(MiscOutflowVolume, INI.ReadString('InputFiles', 'MiscOutflowTimeseries', ''));
    LoadSingleTS(Runoff, INI.ReadString('InputFiles', 'RunoffTimeseries', ''));
  end;

  procedure ReadReservoirMeta;
  var
    StageAreaVolumeFile, StageLeakageFile: string;
    ATimestepOrd: Integer;
  begin
    with AReservoir do
    begin
      BasinArea := INI.ReadFloat('Data', 'SubbasinArea', 0);
      LowerStage := INI.ReadFloat('Data', 'MinimumStage', 0);
      IntakeLevel := INI.ReadFloat('Data', 'IntakeStage', 0);
      SpillLevel := INI.ReadFloat('Data', 'SpillLevel', 0);
      ATimestepOrd := INI.ReadInteger('Data', 'Timestep', 4);
      if ATimestepOrd = 3 then Timestep := tstDaily else
      if ATimestepOrd = 4 then Timestep := tstMonthly else
        Assert(False, 'Time step not implemented');
      TimeSteps := INI.ReadInteger('Data', 'Count', 0);
      DivertSpill := INI.ReadInteger('Data', 'DivertSpill', 1)=1;
      RunoffInMM := INI.ReadInteger('Data', 'RunoffInMM', 0)=1;
      StageAreaVolumeFile := DataDirectory+INI.ReadString('InputFiles', 'hsvb_file', '');
      StageLeakageFile := DataDirectory+INI.ReadString('InputFiles', 'stage-leakage_file', '');
      LoadCurves(StageAreaVolumeFile, StageLeakageFile);
    end;
  end;

  procedure ReadTargets;
  var
    i: Integer;
    s: string;
  begin
    ATargetSet.Clear;
    for i := 0 to 99 do
    begin
      s := 'Target'+IntToStr(i);
      if not INI.SectionExists(s) then Continue;
      ATargetSet.Add(
        INI.ReadInteger(s, 'Priority', 0),
        INI.ReadInteger(s, 'IsDownstream', 0)=1,
        DataDirectory+INI.ReadString(s, 'CurveFilename',''),
        DataDirectory+INI.ReadString(s, 'TimeseriesFilename',''),
        INI.ReadString(s, 'ActualSupplyTimeseries',''));
    end;
    ATargetSet.LoadAll;
  end;

  procedure ClearTimeseries;
  begin
    ResultingOutflow.Clear;
  end;

  procedure AssessCommonPeriod;
  var
    i: Integer;
  begin
    ATimeseriesList := TObjectList.Create(False);
    if Rainfall.Count>0 then ATimeseriesList.Add(Rainfall);
    if Evaporation.Count>0 then ATimeseriesList.Add(Evaporation);
    if MiscInflowVolume.Count>0 then  ATimeseriesList.Add(MiscInflowVolume);
    if MiscOutflowVolume.Count >0 then ATimeseriesList.Add(MiscOutflowVolume);
    if RunInteractive then
      if Runoff.Count>0 then ATimeseriesList.Add(Runoff);
    for i := 0 to ATargetSet.Count-1 do
      with ATargetSet.Items[i] do
        if DemandTimeseries.Count>0 then
          ATimeseriesList.Add(DemandTimeseries);
    ADateTimeList := GetCommonPeriod(ATimeseriesList,0);
  end;

  procedure CheckGaps(ATimestep: TTimestep);
  var
    i: Integer;
  begin
    for i := 0 to ADateTimeList.Count-2 do
      if ADateTimeList[i+1]<>NextStep(ADateTimeList[i], ATimestep) then
        raise Exception.Create('Gaps detected, between dates: '+
          DateToStr(ADateTimeList[i])+' and '+DateToStr(ADateTimeList[i+1]));
  end;

  procedure AssertTimeseries;
  var
    i: Integer;
  begin
    Assert(Rainfall.TimeStep = TimeStep);
    Assert(Evaporation.Timestep = TimeStep);
    Assert(MiscInflowVolume.Timestep = TimeStep);
    Assert(MiscOutflowVolume.Timestep = TimeStep);
    if RunInteractive then
      Assert(Runoff.TimeStep = TimeStep);
    for i := 0 to ATargetSet.Count-1 do
      with ATargetSet.Items[i] do
        Assert(DemandTimeseries.TimeStep = TimeStep);
  end;  

  procedure TimeChecks;
  begin
    AssertTimeseries;
    AssessCommonPeriod;
    CheckGaps(Timestep);
    if ADateTimeList.IndexOf(StartDateTime)<0 then
      raise Exception.Create('Starting date ('+DateTimeToStr(StartDateTime)+
       ') not in time series common period dates');
    if (ADateTimeList.Count - ADateTimeList.IndexOf(StartDateTime))< TimeSteps then
      raise Exception.Create('Requested time ('+IntToStr(TimeSteps)+
      ') steps exceed the time series common period' );
  end;

begin
  Counter := 0 ;      //No TimeStep performed yet
  INI := nil;
  Errors := TStringList.Create ; //Creates the list of error messages
  NumberOfErrorMessages := 0;   //No errors yet
  err := 'Error occured in Initialize method: ' ;
  NilTimeseries;
  AReservoir := nil;
  ATargetSet := nil;
  ADateTimeList := nil;
  ATimeseriesList := nil;
  SavedDecimalSeparator := SysUtils.DecimalSeparator;
  try try
    SysUtils.DecimalSeparator := '.';
    if filepath = '' then filepath := '.\settings.INI';
    if not FileExists(filepath) then
      raise Exception.Create(filepath+': file does not exist');
    INI := TMemIniFile.Create(filepath,TEncoding.UTF8);
    DataDirectory := INI.ReadString('Data', 'DataDirectory', '.')+'/';
    if not DirectoryExists(DataDirectory) then
      raise Exception.Create(DataDirectory+': Non existent directory...');
    OutputDirectory := INI.ReadString('Data', 'OutputDirectory', '.')+'/';
    if not DirectoryExists(OutputDirectory) then
      raise Exception.Create(OutputDirectory+': Non existent directory...');
    ResultingOutflowFile := INI.ReadString('OutputFiles', 'ResultingOutflowTimeseries', '');
    SpillTimeseriesFile := INI.ReadString('OutputFiles', 'SpillTimeseries', '');
    StageTimeseriesFile := INI.ReadString('OutputFiles', 'StageTimeseries', '');
    StorageTimeseriesFile := INI.ReadString('OutputFiles', 'StorageTimeseries', '');
    InitialStage := INI.ReadFloat('Data', 'InitialStage', 0);
    ADateTime := Dates.FormatStrToDateTime('yyyy/mm/dd hh:nn', INI.ReadString(
      'Data', 'StartingDate', '1900/01/01 00:00'));
    StartDateTime := ADateTime;
    AReservoir := TReservoir.Create;
    ATargetSet := TTargetSet.Create;
    ATargetSet.OutputDirectory := OutputDirectory;
    ATimeseriesList := TObjectList.Create(False);
    ReadReservoirMeta;
    ReadTargets;
    CreateTimeseries;
    LoadTimeseries;
    TimeChecks;
    Result := True ;
  except
    on E: Exception do
    begin
      FreeTimeseries;
      NilTimeseries;
      AReservoir.Free;
      AReservoir := nil;
      ATargetSet.Free;
      ATargetSet := nil;
      ADateTimeList.Free;
      ADateTimeList := nil;
      NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
      Errors.Add(err+E.Message) ;
      Result := False ;
    end;
  end;
  finally
    INI.Free;
    ATimeseriesList.Free;
    SysUtils.DecimalSeparator := SavedDecimalSeparator;
  end;
end;

{This function performs a single timestep of the model's simulation
 Storage is the amount of water in the reservoir
 WaterDemand is the flow rate from the reservoir to the next branch of the river
 }

function PerformTimeStep() : Boolean ; cdecl ;
var
  err : ShortString ;
  ARunoff: Real;
  ARunoffType: TRunoffType;
begin
  err := 'Error occured in PerformTimeStep method: ' ;
  try
    if Counter = 0 then
    begin
      Stage := InitialStage;
      StageTimeseries.Add(ADateTime, False, Stage, '', msNew);
    end;
    if not RunInteractive then
      ARunoff := Inflow else
      ARunoff := Runoff[Runoff.IndexOf(ADateTime)].AsFloat;
    ARunoffType := trtVolume;
    if RunInteractive and RunoffInMM then ARunoffType := trtHeight;

    RunSRM(AReservoir, ATargetSet, Stage, ARunoff, ADateTime, Rainfall,
      Evaporation, MiscInflowVolume, MiscOutflowVolume, ARunoffType,
        PrevStorage, Stage, Storage, Outflow, Spill);

    ResultingOutflow.Add(ADateTime, False, Outflow, '', msNew);
    SpillTimeseries.Add(ADateTime, False, Spill, '', msNew);
    if Counter = 0 then
      StorageTimeseries.Add(ADateTime, False, PrevStorage, '', msNew);
    Inc(Counter);
    ADateTime := NextStep(ADateTime, Timestep);
    StageTimeseries.Add(ADateTime, False, Stage, '', msNew);
    StorageTimeseries.Add(ADateTime, False, Storage, '', msNew);
    Result := true ;
  except
    on E: Exception do
    begin
      NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
      Errors.Add(err+E.Message) ;
      Result := false ;
    end;
  end;
end;

{This function is used to close files that the engine used(e.g. input/output files}

function Finish() : Boolean ; cdecl ;
var
  err : ShortString ;
begin
  err := 'Error occured in Finish method: ' ;
 try
  if ResultingOutflowFile<>'' then ResultingOutflow.WriteToFile(OutputDirectory+ResultingOutflowFile);
  if SpillTimeseriesFile<>'' then SpillTimeseries.WriteToFile(OutputDirectory+SpillTimeseriesFile);
  if StorageTimeseriesFile<>'' then StorageTimeseries.WriteToFile(OutputDirectory+StorageTimeseriesFile);
  if StageTimeseriesFile<>'' then StageTimeseries.WriteToFile(OutputDirectory+StageTimeseriesFile);
  ATargetSet.SaveAll;
  Result := true ;
 except
   on E: Exception do
   begin
     NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
     Errors.Add(err+E.Message) ;
     Result := false ;
   end;
 end;
end;

{This function is used for cleaning up/de-allocating memory}

function Dispose() : Boolean ; cdecl ;
var err : ShortString ;
begin
   FreeTimeseries;
   NilTimeseries;
   AReservoir.Free;
   AReservoir := nil;
   ATargetSet.Free;
   ATargetSet := nil;
   ADateTimeList.Free;
   ADateTimeList := nil;
   err := 'Error occured in Dispose method :' ;
 try
   Result := true ;
 except
   on E: Exception do
   begin
     NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
     Errors.Add(err+E.Message) ;
     Result := false ;
   end;
 end;
end;

{This function gives the Storage of the Reservoir}

function GetStorage(var Strg : Double) : Boolean ; cdecl ;
var err : ShortString ;
begin
   err := 'Error occured in GetStorage method' ;
 try
   Strg := Storage *1e6; //Storage in hm3, Return m3
   Result := true ;
 except
   NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting the ModelID
 ShortString is used instead of string in order not to use ShareMem}

function GetModelID(var id : ShortString ) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetModelID method' ;
 try
    id := 'RMM-NTUA' ;
    Result := true;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting a Description for the model
 ShortString is used instead of string in order not to use ShareMem}

function GetModelDescription(var desc : ShortString) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetModelDescription method' ;
 try
    desc := 'RMM-NTUA is a Reservoir Model component' ;
    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting the length of the time step for the simulation}

function GetTimeStepLength(var tstep : Double) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetTimeStepLength method' ;
 try
    if TimeStep = tstMonthly then
        tstep := SubtractDateTime(ADateTime, IncMonth(ADateTime, -1))*24*3600
    else if TimeStep = tstDaily then
        tstep := 24*3600
    else
      Assert(False);
    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting the number of the time steps for the simulation}

function GetNumberOfTimeSteps(var tsteps : Integer) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetNumberOfTimeSteps method' ;
 try
    tsteps := TimeSteps ;
    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{Note: TDateTime is a double precision IEEE floating point value}

{This function is used for getting the time we inputed the data}

function GetInputTime(var InptTime : Double) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetINputTime method' ;
 try
    InptTime := SubtractDateTime(ADateTime , StartDateTime)*3600*24;  //Input time for each time step

    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting the current time of the simulation}

function GetCurrentTime(var CurrTime : Double) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetCurrentTime method' ;
 try
    if Timestep = tstMonthly then
        CurrTime := IncMonth(ADateTime, -1);  //Returns time
    if Timestep = tstDaily then
        CurrTime := SubtractDateTime(ADateTime, 1)
    else
      Assert(False);
    CurrTime := SubtractDateTime(CurrTime,StartDateTime)*3600*24;
    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting the StartDate of the simulation
 Date and Time is displayed as Year/Month/Day Hour:Minute:Seconds
 as given by Now() method of Delphi}

function GetSimulationStartDate(var SimStDate : ShortString) : Boolean ; cdecl ;
var err : ShortString ;
begin
    err := 'Error occured in GetSimulationStartDate method' ;
 try
    SimStDate :=  FormatDateTime('yyyy/mm/dd hh:nn:ss', StartDateTime); //Date and Time of the Simulation start
    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for adding the inflow into the reservoir
 It is essential for OpenMI compatibility
 Inflow is the flow rate to the node we connect the reservoir}

function AddInFlow(infl : Double) : Boolean ; cdecl ;
var
  err : ShortString ;
  StepLength: Double;
begin
    err := 'Error occured in AddInflow method' ;
 try
    GetTimeStepLength(StepLength);
    Inflow := infl*StepLength; // infl in m3/s, Inflow in m3
    Inflow := Inflow/1e6 ;     // Transform m3 to hm3
    Result := true ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This function is used for getting the outflow from the reservoir
 It is essential for OpenMI compatibility}

function GetFlow(var flow : Double) : Boolean ; cdecl ;
var
  err : ShortString ;
  StepLength: Double;
begin
    err := 'Error occured in Getflow method' ;
 try
    GetTimeStepLength(StepLength);
    if Counter <> 0 then    //The Reservoir does not provide water until some is
      flow := Outflow*1e6   //added in its initial storage
    else
      flow := 0 ;
    flow := flow / StepLength; //m3 -> m3/s
    Result := True ;
 except
    NumberOfErrorMessages := NumberOfErrorMessages + 1 ;
    Errors.Add(err) ;
    Result := false ;
 end;
end;

{This procedure is used for handling errors}

procedure GetMessage(var n : Integer ; var MessageFromCore : ShortString) cdecl ;
begin
     MessageFromCore := Errors[n] ;
end;

{This function is used for counting the errors}

function GetNumberOfMessages() : Integer ; cdecl ;
begin
     Result := NumberOfErrorMessages ;
end;

{This function runs the whole simulation
 The pascal attribute is for passing parameters correctly}

function RunSimulation: PChar; pascal;
begin
  Result := RunSimulationArgs('');
end;

function RunSimulationArgs(filepath : ShortString): PChar; pascal;
begin
  try
    RunInteractive := True;
    Result := '';
    Initialize(filepath);
    repeat
      if NumberOfErrorMessages>0 then Break;
      if NumberOfErrorMessages=0 then PerformTimeStep;
    until Counter>=TimeSteps;
    if NumberOfErrorMessages=0 then Finish();
    Dispose();
    if NumberOfErrorMessages>0 then
      Result := PChar(Errors.Text);
  finally
    RunInteractive := False;
  end;
end;

end.
